/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.jambalaya;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import ca.uvic.csr.shrimp.AttributeConstants;
import ca.uvic.csr.shrimp.LayoutConstants;
import ca.uvic.csr.shrimp.ShrimpConstants;
import ca.uvic.csr.shrimp.VisVarConstants;
import ca.uvic.csr.shrimp.AttrToVisVarBean.AttrToVisVarBean;
import ca.uvic.csr.shrimp.DisplayBean.DisplayConstants;
import ca.uvic.csr.shrimp.DisplayBean.ArcStyle.ArcStyle;
import ca.uvic.csr.shrimp.DisplayBean.ArcStyle.StraightDottedLineArcStyle;
import ca.uvic.csr.shrimp.DisplayBean.ArcStyle.StraightSolidLineArcStyle;
import ca.uvic.csr.shrimp.DisplayBean.NodeShape.EllipseNodeShape;
import ca.uvic.csr.shrimp.ShrimpApplication.BeanNotFoundException;
import ca.uvic.csr.shrimp.ShrimpApplication.ShrimpToolNotFoundException;
import ca.uvic.csr.shrimp.ShrimpProject.AbstractShrimpProject;
import ca.uvic.csr.shrimp.ShrimpProject.ShrimpProject;
import ca.uvic.csr.shrimp.gui.ActionManager.ActionManager;
import ca.uvic.csr.shrimp.gui.ActionManager.MenuAction;
import ca.uvic.csr.shrimp.gui.ActionManager.ShrimpAction;
import ca.uvic.csr.shrimp.gui.ShrimpView.ShrimpView;
import ca.uvic.csr.shrimp.gui.quickview.QuickViewAction;
import ca.uvic.csr.shrimp.jambalaya.ProtegeDataBean.ProtegeDataBean;
import ca.uvic.csr.shrimp.jambalaya.ProtegeDataBean.ProtegeOWLDataBean;
import ca.uvic.csr.shrimp.usercontrols.NestedTreemapViewAction;
import edu.stanford.smi.protege.model.Project;
import edu.stanford.smi.protege.resource.Icons;
import edu.stanford.smi.protege.ui.DisplayUtilities;

/**
 * JambalayaProject represents a single project in Jambalaya.
 * 
 * @author Nasir Rather, Rob Lintern, Chris Callendar
 */
public class JambalayaProject extends AbstractShrimpProject {
	
	public static final String EXT_OWL = "owl";
	public static final String PROJECT_TYPE_OWL = "OWL";
	public static final String ACTION_NAME_CHANGE_ROOT_CLSES = "Change Root Classes...";
	
    public static final Color CLASS_COLOR = new Color (245, 224, 171);
    public static final Color META_CLASS_COLOR = new Color (249, 238, 211);
	public static final Color SYSTEM_CLASS_COLOR = new Color (249, 238, 211);
	public static final Color INSTANCE_COLOR = new Color (223, 214, 223);
    public static final Color DIRECT_SUBCLASS_COLOR = Color.BLUE;
    public static final Color DIRECT_INSTANCE_COLOR = Color.RED;
	
	private Project protegeProject;
	private boolean isOwlProject = false;
    private boolean attrToVisVarBeanSetup = false;

	/**
	 * Create JambalayaProject.
	 * @param application The application that contains this project.
	 * @param protegeProject The Protege project that this Jambalaya project should be attached to. 
	 * @param sourceURI The project URI
	 */
	public JambalayaProject(JambalayaApplication application, Project protegeProject, URI sourceURI) {
		super(application, determineProjectType(protegeProject));
		this.protegeProject = protegeProject;
		this.projectURI = sourceURI;
		if (projectURI == null) {
			projectURI = protegeProject.getProjectURI();
			if (projectURI == null) {
				projectURI = protegeProject.getLoadingURI();
			}
			if (projectURI == null) {
			    try {
	                projectURI = new URI("file:///unknown");
	            } catch (URISyntaxException e) {}
			}
		}
		this.title = getProjectTitle(protegeProject);
		
		projectActions = new Vector();
		
        // filename where the properties of this project will be stored
        String appName = application.getName().replace(' ', '_');
        if (projectURI.getScheme().equals("file")) {
        	File pprjFile = new File(projectURI.getPath());
        	File propFile = new File(pprjFile.getParentFile(), title + "_" + appName + ".properties");
            propertiesFileName =  propFile.getAbsolutePath();
        } else {
            try {
                propertiesFileName = System.getProperty("java.io.tmpdir");
                if (propertiesFileName != null) {
                    propertiesFileName += title + "_" + appName  + ".properties";
                }
            } catch (SecurityException e){
                // thrown if don't have proper security priviledges
                propertiesFileName = null;
            }
        }

		loadProperties();

		isOwlProject = JambalayaProjectPlugin.isOwlProject(protegeProject);
		
        // create data bean
		ProtegeDataBean dataBean = isOwlProject ? new ProtegeOWLDataBean(protegeProject) : new ProtegeDataBean(protegeProject);
		dataBean.setDefaultRootClses();
		
		// add databean to this project
		addBean(ShrimpProject.DATA_BEAN, dataBean);
		
		createQuickViewActions();
	}
	
	/**
	 * Determines a title for the given project.
	 * If the project is null, the "Null Project" is returned.
	 * If the project has a name then that is returned.
	 * Otherwise the project URI or loading URI is used to determine a name.
	 * Defaults to "Unknown".
	 * @param project
	 * @return String title - shouldn't be null
	 */
	public static String getProjectTitle(Project project) {
		String title = null;
		if (project != null) {
	        if (project.getProjectName() != null) {
	            title = project.getProjectName();
	        } else {
	            // remote and unsaved projects have no name...we'll use the last part of the URI for a name
	        	URI uri = project.getProjectURI();
	        	if (uri == null) {
	        		uri = project.getLoadingURI();
		        	if (uri == null) {
		        		 try {
		                     uri = new URI("file:///Unknown");
		                 } catch (URISyntaxException e) {}
		        	}
	        	}
	            title = uri.toString();
	            int index = title.lastIndexOf('/');
	            if (index != -1) {
	                title = title.substring(index + 1);
	            }
	        }			
		} else {
			title = "Null Project";
		}
		if (title == null) {
			title = "Unknown";
		}
		return title;
	}
	
	/**
	 * Determins the project type - either OWL or Protege.
	 * @param project
	 * @return String either OWL or Protege.
	 */
	private static String determineProjectType(Project project) {
		String type;
		if (JambalayaProjectPlugin.isOwlProject(project)) {
			type = PROJECT_TYPE_OWL;
		} else {
			type = PROJECT_TYPE_PROTEGE;
		}
		return type;
	}

	/* (non-Javadoc)
     * @see ca.uvic.csr.shrimp.ShrimpProject.AbstractShrimpProject#getBean(java.lang.String)
     */
    public Object getBean(String name) throws BeanNotFoundException {
        Object bean = super.getBean(name);
        if (bean != null && name.equals(ShrimpProject.ATTR_TO_VIS_VAR_BEAN) && !attrToVisVarBeanSetup) {
            setupAttrToVisVarBean((AttrToVisVarBean)bean);
        }
        return bean;
    }
	
	private void setupAttrToVisVarBean (AttrToVisVarBean attrToVisVarBean) {
		boolean firingEvents = attrToVisVarBean.isFiringEvents();
		attrToVisVarBean.setFiringEvents(false);
		
		//set some default colours for classes and instances
		//regular protege colours
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeDataBean.CLASS_ART_TYPE, CLASS_COLOR);
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeDataBean.INSTANCE_ART_TYPE, INSTANCE_COLOR);
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeDataBean.META_CLASS_ART_TYPE, META_CLASS_COLOR);
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeDataBean.SYSTEM_CLASS_ART_TYPE, SYSTEM_CLASS_COLOR);
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeDataBean.SYSTEM_META_CLASS_ART_TYPE, SYSTEM_CLASS_COLOR);
		
		//owl plugin colours
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeOWLDataBean.PRIMITIVE_CLASS_TYPE, CLASS_COLOR);		
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeOWLDataBean.DEFINED_CLASS_TYPE, CLASS_COLOR);
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeOWLDataBean.INDIVIDUAL_TYPE, INSTANCE_COLOR);
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeOWLDataBean.LOGICAL_OPERATION_TYPE, new Color (249, 241, 168));
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_COLOR, ProtegeOWLDataBean.RESTRICTION_TYPE, new Color (249, 241, 168));
		
		//make restrictions round
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_SHAPE, ProtegeOWLDataBean.RESTRICTION_TYPE, new EllipseNodeShape ());
		
		//make logical expressions round
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_SHAPE, ProtegeOWLDataBean.LOGICAL_OPERATION_TYPE, new EllipseNodeShape ());

		// setup some default icons for node types
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_ICON, ProtegeDataBean.CLASS_ART_TYPE, Icons.getClsIcon());
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_ICON, ProtegeDataBean.INSTANCE_ART_TYPE, Icons.getInstanceIcon());
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_ICON, ProtegeDataBean.META_CLASS_ART_TYPE, Icons.getClsIcon(true, false, false, false));
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_ICON, ProtegeDataBean.SYSTEM_CLASS_ART_TYPE, Icons.getClsIcon(false, false, true, false));
		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, VisVarConstants.VIS_VAR_NODE_ICON, ProtegeDataBean.SYSTEM_META_CLASS_ART_TYPE, Icons.getClsIcon(true, false, true, false));
		
        try { 
            OWLSpecificUtils.setOWLIcons(attrToVisVarBean);
        } catch (Throwable e) {
            // owl plugin is not installed
        }
		
		// give specific relationship types some default arc styles
		ArcStyle straightDottedArcStyle = new StraightDottedLineArcStyle();
		ArcStyle straightSolidArcStyle = new StraightSolidLineArcStyle();
		
        attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_REL_TYPE, VisVarConstants.VIS_VAR_ARC_COLOR, ProtegeDataBean.DIRECT_SUBCLASS_SLOT_TYPE, DIRECT_SUBCLASS_COLOR);     
        attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_REL_TYPE, VisVarConstants.VIS_VAR_ARC_COLOR, ProtegeDataBean.DIRECT_INSTANCE_SLOT_TYPE, DIRECT_INSTANCE_COLOR);     
		
        try {
            ProtegeDataBean dataBean = (ProtegeDataBean) getBean(ShrimpProject.DATA_BEAN);
            Vector allRelTypes = dataBean.getRelationshipTypes(false, true);
            for (Iterator iter = allRelTypes.iterator(); iter.hasNext();) {
            	String relType = (String) iter.next();
            	String groupName = dataBean.getDefaultGroupForRelationshipType(relType);
            	boolean isSystem = groupName.equals(ProtegeDataBean.GROUP_NAME_SYSTEM);
            	boolean isTemplateAllowed = relType.indexOf(ProtegeDataBean.SUFFIX_SLOT_TEMPLATE_ALLOWED_CLASS) != -1; 
            	boolean isTemplateValue = relType.indexOf(ProtegeDataBean.SUFFIX_SLOT_TEMPLATE_VALUE) != -1;
            	boolean isDomainRange = isOwlProject && relType.indexOf(ProtegeOWLDataBean.SUFFIX_DOMAIN_RANGE) != -1;
            	boolean isInherited = isOwlProject && relType.indexOf(ProtegeOWLDataBean.SUFFIX_INHERITED) != -1;
            	if (isSystem) {
            		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_REL_TYPE, VisVarConstants.VIS_VAR_ARC_STYLE, relType, straightSolidArcStyle);
            	} else if (isTemplateAllowed || isTemplateValue || isDomainRange || isInherited) {
            		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_REL_TYPE, VisVarConstants.VIS_VAR_ARC_STYLE, relType, straightDottedArcStyle);
            	} else {
            		attrToVisVarBean.setDefaultNominalVisualVariableValue(AttributeConstants.NOM_ATTR_REL_TYPE, VisVarConstants.VIS_VAR_ARC_STYLE, relType, straightSolidArcStyle);
            	}
            }
        } catch (BeanNotFoundException e) {
            e.printStackTrace();
        }
		
		attrToVisVarBean.setFiringEvents(firingEvents);			
		attrToVisVarBeanSetup = true;
	    
	}
	
	public void createDefaultQuickViewActions() {
		super.createDefaultQuickViewActions();
		
		if (isOwlProject) {
			createDefaultOWLQuickViewActions();
		} else {
			createDefaultProtegeQuickViewActions();
		}
		
		// @tag Jambalaya(QueryViewQuickView) : add the query view quick view
		/*
		action = new NeighbourhoodAction(JambalayaApplication.QUICK_VIEW_QUERY_VIEW, ResourceHandler.getIcon("icon_quick_view_qv.gif"));
		addProjectAction(action, ShrimpConstants.MENU_QUICK_VIEWS, ShrimpConstants.GROUP_A, position++);
		*/   
	}

	protected int createDefaultOWLQuickViewActions() {
		ActionManager actionManager = getActionManager();
		boolean firingEvents = actionManager.getFiringEvents();
		actionManager.setFiringEvents(false);

		int position = 1;
		
		// Flat Class & Individual Tree
		List nodeTypesOfInterest = new ArrayList();
		nodeTypesOfInterest.add(ProtegeOWLDataBean.DEFINED_CLASS_TYPE);
		nodeTypesOfInterest.add(ProtegeOWLDataBean.PRIMITIVE_CLASS_TYPE);
		nodeTypesOfInterest.add(ProtegeOWLDataBean.INDIVIDUAL_TYPE);
		nodeTypesOfInterest.add(ProtegeOWLDataBean.RDFS_CLASS_TYPE);
		List arcTypesOfInterest = new ArrayList();
		arcTypesOfInterest.add(ProtegeDataBean.DIRECT_SUBCLASS_SLOT_TYPE);
		arcTypesOfInterest.add(ProtegeDataBean.DIRECT_INSTANCE_SLOT_TYPE);
		QuickViewAction action = QuickViewAction.createFlatView(JambalayaApplication.QUICK_VIEW_CLASS_INDIVIDUAL_TREE, this, nodeTypesOfInterest,
				arcTypesOfInterest, LayoutConstants.LAYOUT_TREE_VERTICAL, DisplayConstants.LABEL_MODE_FIXED);
		action.setIconFilename("icon_quick_view_cih.gif");
		addProjectAction(action, ShrimpConstants.MENU_QUICK_VIEWS, ShrimpConstants.GROUP_A, position++);

		// Flat Class Tree
		nodeTypesOfInterest = new ArrayList();
		nodeTypesOfInterest.add(ProtegeOWLDataBean.DEFINED_CLASS_TYPE);
		nodeTypesOfInterest.add(ProtegeOWLDataBean.PRIMITIVE_CLASS_TYPE);
		nodeTypesOfInterest.add(ProtegeOWLDataBean.RDFS_CLASS_TYPE);
		arcTypesOfInterest = new ArrayList();
		arcTypesOfInterest.add(ProtegeDataBean.DIRECT_SUBCLASS_SLOT_TYPE);
		action = QuickViewAction.createFlatView(JambalayaApplication.QUICK_VIEW_CLASS_TREE, this, nodeTypesOfInterest,
				arcTypesOfInterest, LayoutConstants.LAYOUT_TREE_VERTICAL, DisplayConstants.LABEL_MODE_FIXED);
		action.setIconFilename("icon_quick_view_ch.gif");
		addProjectAction(action, ShrimpConstants.MENU_QUICK_VIEWS, ShrimpConstants.GROUP_A, position++);

		// Flat Domain-Range Spring
		nodeTypesOfInterest = new ArrayList();
		nodeTypesOfInterest.add(ProtegeOWLDataBean.DEFINED_CLASS_TYPE);
		nodeTypesOfInterest.add(ProtegeOWLDataBean.PRIMITIVE_CLASS_TYPE);
		nodeTypesOfInterest.add(ProtegeOWLDataBean.RDFS_CLASS_TYPE);
		// nodeTypesOfInterest.add(ProtegeOWLDataBean.LOGICAL_OPERATION_TYPE);
		arcTypesOfInterest = new ArrayList();
		try {
			ProtegeDataBean dataBean = (ProtegeDataBean) getBean(ShrimpProject.DATA_BEAN);
			Vector allRelTypes = dataBean.getRelationshipTypes(false, true);
			for (Iterator iter = allRelTypes.iterator(); iter.hasNext();) {
				String relType = (String) iter.next();
				String groupName = dataBean.getDefaultGroupForRelationshipType(relType);
				if (groupName.equals(ProtegeOWLDataBean.GROUP_NAME_DOMAIN_RANGE)) {
					arcTypesOfInterest.add(relType);
				}
			}
		} catch (BeanNotFoundException e) {
			e.printStackTrace();
		}
		action = QuickViewAction.createFlatView(JambalayaApplication.QUICK_VIEW_DOMAIN_RANGE, this, nodeTypesOfInterest,
						arcTypesOfInterest, LayoutConstants.LAYOUT_SPRING, DisplayConstants.LABEL_MODE_FIXED);
		action.setIconFilename("icon_quick_view_fio.gif");
		addProjectAction(action, ShrimpConstants.MENU_QUICK_VIEWS, ShrimpConstants.GROUP_A, position++);

		action = new NestedTreemapViewAction(this);
        addProjectAction(action, ShrimpConstants.MENU_QUICK_VIEWS, ShrimpConstants.GROUP_A, position++);
		
		actionManager.setFiringEvents(firingEvents);
        
		return position;
	}
	
	/**
	 * @see ca.uvic.csr.shrimp.ShrimpProject.AbstractShrimpProject#createProjectActions()
	 */
	public void createProjectActions() {
		super.createProjectActions();
		
		ActionManager actionManager = getActionManager();
		boolean firingEvents = actionManager.getFiringEvents();
		actionManager.setFiringEvents(false);
		
		ShrimpAction action = new MenuAction(ACTION_NAME_CHANGE_ROOT_CLSES) {
			public void actionPerformed(ActionEvent e) {
				changeRootClasses();
			}
		};
		addProjectAction(action, ShrimpConstants.MENU_EDIT, ShrimpConstants.GROUP_D, 1);
		
		actionManager.setFiringEvents(firingEvents);
	}
	
	private void changeRootClasses() {
		try {
			ShrimpView shrimpView = (ShrimpView) getTool(ShrimpProject.SHRIMP_VIEW);
			ProtegeDataBean dataBean = (ProtegeDataBean)getBean(ShrimpProject.DATA_BEAN);

			Collection clses = DisplayUtilities.pickClses(getApplication().getParentFrame(), dataBean.getKnowledgeBase());
			if (!clses.isEmpty()) {
				HashSet clsesSet = new HashSet(clses);
				dataBean.setRootClses(clsesSet);	
				// if the jambalaya tab is open then set the roots of the classes tree
				JambalayaTab jambalayaTab = JambalayaProjectPlugin.getJambalayaTab();
				if (jambalayaTab != null) {
				    jambalayaTab.setRootClses(clsesSet);
				}
				refresh();
				shrimpView.addDefaultRootNodes(true);
			}
		} catch (BeanNotFoundException e1) {
			//e1.printStackTrace();
		} catch (ShrimpToolNotFoundException e1) {
			//e1.printStackTrace();
		} 
	}

	/** 
	 * @see ca.uvic.csr.shrimp.ShrimpProject.AbstractShrimpProject#refresh()
	 */
	public void refresh() {
	    super.refresh();
        JambalayaProjectPlugin.getJambalayaProjectPlugin().jambalayaProjectRefreshed();
	}

	/**
	 * @see ca.uvic.csr.shrimp.ShrimpProject.AbstractShrimpProject#disposeProject()
	 */
	public void disposeProject() {
		super.disposeProject();
		// no need to dispose the protege project, it gets disposed by the ProjectManager 
		// after JambalayaProjectPlugin.beforeClose(project) is called.
		protegeProject = null;
	}

    public Project getProtegeProject() {
        return protegeProject;
    }
}
