package org.eclipse.mylar.zest.layouts.algorithms;

import java.util.Arrays;
import java.util.Comparator;

import org.eclipse.mylar.zest.layouts.constraints.EntityPriorityConstraint;
import org.eclipse.mylar.zest.layouts.dataStructures.InternalNode;
import org.eclipse.mylar.zest.layouts.dataStructures.InternalRelationship;

import edu.umd.cs.treemap.MapLayout;
import edu.umd.cs.treemap.Mappable;
import edu.umd.cs.treemap.Rect;
import edu.umd.cs.treemap.SimpleMapModel;
import edu.umd.cs.treemap.StripTreemap;

/**
 * 
 * @author Ian Bull
 */
public class TreemapLayout extends AbstractLayoutAlgorithm {
	
	public static int BORDER = 5;

	public TreemapLayout(int styles) {
		super(styles);
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.mylar.zest.layouts.algorithms.AbstractLayoutAlgorithm#applyLayoutInternal(org.eclipse.mylar.zest.layouts.dataStructures.InternalNode[], org.eclipse.mylar.zest.layouts.dataStructures.InternalRelationship[], double, double, double, double)
	 */
	protected void applyLayoutInternal(InternalNode[] entitiesToLayout,
			InternalRelationship[] relationshipsToConsider, double boundsX,
			double boundsY, double boundsWidth, double boundsHeight) {
		
		MapLayout layout = new StripTreemap();
		SimpleMapModel mapModel = new SimpleMapModel();
		TreemapNode[] mappableNodes = createMappableNodes(entitiesToLayout);
		setOrder( mappableNodes );
		mapModel.setItems(mappableNodes);
		layout.layout(mapModel, new Rect(boundsX, boundsY, boundsWidth, boundsHeight));
		setLayoutLocation(mappableNodes);
	}
	

	/**
	 * This comparator orders by a Treemap Node's size
	 * @author Ian Bull
	 */
	class OrderComparator implements Comparator {

		/*
		 * (non-Javadoc)
		 * @see java.util.Comparator#compare(T, T)
		 */
		public int compare(Object arg0, Object arg1) {
			TreemapNode o1 = (TreemapNode) arg0;
			TreemapNode o2 = (TreemapNode) arg1;
			return (int)( o2.getSize() - o1.getSize() );
		} 
	}
	
	/**
	 * Orders the nodes in a graph by the node's size property
	 * @param mappableNodes
	 */
	private void setOrder(TreemapNode[] mappableNodes) {
		Arrays.sort(mappableNodes, new OrderComparator());	
		for (int i = 0; i < mappableNodes.length; i++) {
			mappableNodes[i].setOrder(i);
		}
	}

	/**
	 * Set the layout location and uses a border between nodes
	 * @param nodes
	 */
	private void setLayoutLocation(TreemapNode[] nodes) {
		for (int i = 0; i < nodes.length; i++) {
			TreemapNode node = nodes[i];
			Rect bounds = node.getBounds();			
			double width = Math.max ((int) bounds.w - BORDER, 5);
			double height = Math.max ((int) bounds.h - BORDER, 5);
			node.getInternalNode().setInternalLocation(bounds.x, bounds.y);
			node.getInternalNode().setInternalSize(width, height);
		}
	}
	
	
	/**
	 * Creates nodes that can be used with the TreeMap jar
	 * @param nodes
	 * @return
	 */
	private TreemapNode[] createMappableNodes( InternalNode[] nodes ) {
		TreemapNode[] treemapNodes = new TreemapNode[ nodes.length ];
		EntityPriorityConstraint constraint = new EntityPriorityConstraint();
		for (int i = 0; i < nodes.length; i++) {
			constraint.clear();
			nodes[i].getLayoutEntity().populateLayoutConstraint(constraint);
			TreemapNode node = new TreemapNode( nodes[i], constraint.priority );
			treemapNodes[i]= node;
		}
		return treemapNodes;
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.mylar.zest.layouts.algorithms.AbstractLayoutAlgorithm#getCurrentLayoutStep()
	 */
	protected int getCurrentLayoutStep() {
		return 0;
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.mylar.zest.layouts.algorithms.AbstractLayoutAlgorithm#getTotalNumberOfLayoutSteps()
	 */
	protected int getTotalNumberOfLayoutSteps() {
		return 0;
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.mylar.zest.layouts.algorithms.AbstractLayoutAlgorithm#isValidConfiguration(boolean, boolean)
	 */
	protected boolean isValidConfiguration(boolean asynchronous,
			boolean continuous) {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.mylar.zest.layouts.algorithms.AbstractLayoutAlgorithm#postLayoutAlgorithm(org.eclipse.mylar.zest.layouts.dataStructures.InternalNode[], org.eclipse.mylar.zest.layouts.dataStructures.InternalRelationship[])
	 */
	protected void postLayoutAlgorithm(InternalNode[] entitiesToLayout,
			InternalRelationship[] relationshipsToConsider) {
		updateLayoutLocations(entitiesToLayout);
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.mylar.zest.layouts.algorithms.AbstractLayoutAlgorithm#preLayoutAlgorithm(org.eclipse.mylar.zest.layouts.dataStructures.InternalNode[], org.eclipse.mylar.zest.layouts.dataStructures.InternalRelationship[], double, double, double, double)
	 */
	protected void preLayoutAlgorithm(InternalNode[] entitiesToLayout,
			InternalRelationship[] relationshipsToConsider, double x, double y,
			double width, double height) {

	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.mylar.zest.layouts.algorithms.AbstractLayoutAlgorithm#setLayoutArea(double, double, double, double)
	 */
	public void setLayoutArea(double x, double y, double width, double height) {
	}

}

/**
 * A node that can be used with the treemap layout
 * @author Ian Bull
 *
 */
class TreemapNode implements Mappable {

	private InternalNode internalNode;
	private Rect rectangle = new Rect();
	private double size;
	private int order = 0;
	
	/**
	 * Creates a node with a given size.  The size
	 * is the constraint used in treemap
	 * @param node
	 * @param size
	 */
	public TreemapNode(InternalNode node, double size) {
		this.internalNode = node;
		this.size = size;
	}

	/*
	 * (non-Javadoc)
	 * @see edu.umd.cs.treemap.Mappable#getBounds()
	 */
	public Rect getBounds() {
		return rectangle;
	}

	/*
	 * (non-Javadoc)
	 * @see edu.umd.cs.treemap.Mappable#getDepth()
	 */
	public int getDepth() {
		return 0;
	}

	/*
	 * (non-Javadoc)
	 * @see edu.umd.cs.treemap.Mappable#getOrder()
	 */
	public int getOrder() {
		return this.order;
	}

	/*
	 * (non-Javadoc)
	 * @see edu.umd.cs.treemap.Mappable#getSize()
	 */
	public double getSize() {
		return size;
	}

	/*
	 * (non-Javadoc)
	 * @see edu.umd.cs.treemap.Mappable#setBounds(edu.umd.cs.treemap.Rect)
	 */
	public void setBounds(Rect arg0) {
		this.rectangle = arg0.copy();
	}

	/*
	 * (non-Javadoc)
	 * @see edu.umd.cs.treemap.Mappable#setBounds(double, double, double, double)
	 */
	public void setBounds(double arg0, double arg1, double arg2, double arg3) {
		this.rectangle.x = arg0;
		this.rectangle.y = arg1;
		this.rectangle.w = arg2;
		this.rectangle.h = arg3;
	}

	/*
	 * (non-Javadoc)
	 * @see edu.umd.cs.treemap.Mappable#setDepth(int)
	 */
	public void setDepth(int arg0) {

	}

	/*
	 * (non-Javadoc)
	 * @see edu.umd.cs.treemap.Mappable#setOrder(int)
	 */
	public void setOrder(int arg0) {
		this.order = arg0;
	}

	/*
	 * (non-Javadoc)
	 * @see edu.umd.cs.treemap.Mappable#setSize(double)
	 */
	public void setSize(double arg0) {
		this.size = arg0;
	}


	/**
	 * Gets the internal node associated with this treemap layout node
	 * @return
	 */
	public InternalNode getInternalNode() {
		return internalNode;
	}
}